/*
 * Notification support.
 *
 * Copyright (C) 2021, 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <map>
#include <mutex>
#include <set>
#include <vector>

#include <systypes/base.h>



/* Forward declaration. */

class NotifierPeerEndpoint;

/* Collection data types. */

typedef std::set<l4_cap_idx_t> NotifierSet;
typedef std::map<l4_cap_idx_t, unsigned int> NotifierEndpointMap;

/* Notification endpoint details. */

class NotificationEndpoint
{
public:
    NotifierSet notifiers;

    /* Notification flag mask, deferred flags. */

    notify_flags_t flags = 0, deferred_flags = 0;

    /* Deferred values. */

    notify_values_t deferred_values = NOTIFY_VALUES_NULL;
};

typedef std::vector<NotificationEndpoint> NotificationEndpointList;



/* Notification support, allowing pipe or file users to receive events due to
   activity on those objects. */

class NotificationSupport
{
protected:
    std::mutex _lock;

    /* Endpoint count. */

    unsigned int _min_endpoints;

    /* Notification endpoint details. */

    NotificationEndpointList _endpoints;

    /* Notifier endpoint details. */

    NotifierEndpointMap _subscribers;

    /* Deferred notifications for new endpoints. */

    notify_flags_t _deferred_flags = 0;
    notify_values_t _deferred_values = NOTIFY_VALUES_NULL;

    virtual void release_notifiers();

    virtual long _get_endpoint(l4_cap_idx_t notifier, bool remove,
                               unsigned int *endpoint);

    virtual void _notify(unsigned int endpoint, notify_flags_t flags,
                         notify_values_t values);

    virtual void _notify_others(unsigned int endpoint, notify_flags_t flags,
                                notify_values_t values);

    virtual long _subscribe(unsigned int endpoint, l4_cap_idx_t notifier,
                            notify_flags_t flags);

    virtual void _unsubscribe(l4_cap_idx_t notifier);

public:
    explicit NotificationSupport(unsigned int endpoints=0);

    virtual ~NotificationSupport();

    virtual void notify(unsigned int endpoint, notify_flags_t flags,
                        notify_values_t values);

    virtual void notify_all(notify_flags_t flags, notify_values_t values);

    virtual void notify_others(unsigned int endpoint, notify_flags_t flags,
                               notify_values_t values);

    virtual void notify_others(l4_cap_idx_t notifier, notify_flags_t flags,
                               notify_values_t values);

    virtual long subscribe(l4_cap_idx_t notifier, notify_flags_t flags);

    virtual long subscribe(unsigned int endpoint, l4_cap_idx_t notifier,
                           notify_flags_t flags);

    virtual void unsubscribe(l4_cap_idx_t notifier);
};

// vim: tabstop=4 expandtab shiftwidth=4
